import argparse
import json
import os
import random
import sys
from datetime import datetime, timedelta, timezone
from threading import Thread
from typing import Callable, Tuple

import boto3
import sqlalchemy
import urwid
from dateutil import tz
from sqlalchemy.sql import or_

from seesaw import storage
from seesaw.log_factory import polling_worker_entry
from seesaw.config import LOG_DIR, SETTINGS, USER_DIR, logger

log = logger(__file__)
localtz = tz.tzlocal()


def random_hex():
    return "".join(random.choice("abcdef1234567890") for _ in range(10))


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--show-log-file",
        action="store_true",
        help="display the path to this app's log file",
    )
    args = parser.parse_args()
    log.info("launched", args=vars(args))

    if args.show_log_file:
        print(str(LOG_DIR / "logs.json"))
        return

    try:
        loop()
    except KeyboardInterrupt:
        sys.exit(3)


def d_widget(sawmill):
    def show_or_exit(key):
        if key in ("q", "Q"):
            raise urwid.ExitMainLoop()
        elif key in ("v", "V"):
            sawmill.toggle_date()
            sawmill.refresh_logs()
        elif key in ("r", "R"):
            sawmill.refresh_logs()

    return show_or_exit


class LogPollingManager:
    def __init__(self, session: sqlalchemy.orm.Session) -> None:
        self.pollers = {}
        self.session = session

    def pop_group(self, region, log_group_name):
        try:
            self.active.remove((region, log_group_name))
        except KeyError:
            pass

    def push_group(self, region, log_group_name):
        # spin down all queries except ones for the current region
        storage.LiveReader.destroy_other_region_readers(self.session, region)

        thread_id = (region, log_group_name)
        if (
            storage.LiveReader.exists_for(self.session, region, log_group_name)
            and self.pollers.get(thread_id)
            and self.pollers[thread_id].is_alive()
        ):
            log.info("Thread was already alive", args=thread_id)
            return

        log.info(
            "Pushing new poller for log group",
            new=thread_id,
        )
        new_thread = storage.LiveReader(
            pid=os.getpid(), region=region, log_group=log_group_name
        )
        self.session.add(new_thread)
        self.session.commit()
        self.pollers[thread_id] = Thread(
            daemon=True,
            target=polling_worker_entry,
            args=(region, log_group_name),
        )
        self.pollers[thread_id].start()


class LogReadingPane:
    def __init__(self) -> None:
        self.messages = urwid.SimpleFocusListWalker([])
        self.viewer = urwid.ListBox(self.messages)
        self.db = storage.connect()
        self.poll_manager = LogPollingManager(self.db)
        self.verbose_dates = False

        for old_evt in (
            self.db.query(storage.Event)
            .filter(
                storage.Event.timestamp
                < (datetime.now(tz=timezone.utc) - timedelta(hours=8)).isoformat()
            )
            .all()
        ):
            log.info("Deleting", e=old_evt.timestamp)
            self.db.delete(old_evt)
        self.db.commit()

        self.region_menu = TargetSelect()
        self.top = urwid.Columns([(90, self.region_menu), self.viewer], dividechars=2)

    def refresh_logs(self):
        region, groups = self.region_menu.get_selected_groups()
        for g in groups:
            self.poll_manager.push_group(region, g)
        try:
            log.info("querying local storage", region=region, groups=groups)
            self.messages.clear()
            for event in sorted(
                self.db.query(storage.Event)
                .filter(storage.Event.region == region)
                .filter(
                    storage.Event.timestamp
                    > (datetime.now(tz=timezone.utc) - timedelta(hours=8)).isoformat()
                )
                .filter(
                    or_(storage.Event.log_group == g for g in groups)
                )  # expands to the equivalent of `if Event.log_group in groups`
                .order_by(storage.Event.timestamp.desc())
                .limit(500)
                .all(),
                key=lambda e: datetime.fromisoformat(e.timestamp),
            ):
                log.debug("Found event", msg=event.message, gid=event.gid)
                self.new_message(event)
        except:
            log.exception("Error rendering log messages")

    def new_message(self, event):
        self.messages.append(urwid.Text(self.format_event(event)))
        self.viewer.focus_position = len(self.messages) - 1

    def toggle_date(self):
        self.verbose_dates = not self.verbose_dates

    def timestamp(self, ts):
        if self.verbose_dates:
            return (
                "timestamp",
                datetime.fromisoformat(ts)
                .astimezone(localtz)
                .strftime("%Y-%m-%d %H:%M:%S"),
            )
        return (
            "timestamp",
            datetime.fromisoformat(ts).astimezone(localtz).strftime("%H:%M:%S"),
        )

    def format_event(self, event):
        try:
            return [
                self.timestamp(event.timestamp),
                " ",
                json.dumps(json.loads(event.message), indent=2, sort_keys=True),
            ]
        except:
            return [
                self.timestamp(event.timestamp),
                " ",
                event.message,
            ]


class TargetSelect(urwid.WidgetWrap):
    def __init__(self):
        self.regions = [r[0] for r in initial_options()]
        self.options = []

        group_select = GroupSelect(get_selected_region=self.get_state)
        self._selected_groups = group_select.get_state

        def update_region_selection(caller_widget, new_state):
            if not new_state:
                log.info("User region deselected", region=caller_widget._label._text)
                return
            log.info("User region selected", region=caller_widget._label._text)
            group_select.change_region(caller_widget._label._text)

        widgets = [
            urwid.RadioButton(
                self.options,
                region_name,
                state=(region_name == SETTINGS["presets"]["default"]["region"]),
                on_state_change=update_region_selection,
            )
            for region_name in self.regions
        ]
        region_list = urwid.ListBox(widgets)
        display_widget = urwid.Columns(
            [(18, region_list), (70, group_select)], dividechars=2
        )
        urwid.WidgetWrap.__init__(self, display_widget)

    def get_selected_groups(self) -> Tuple[str, set]:
        return self.get_state(), self._selected_groups()

    def get_state(self):
        for o in self.options:
            if o.get_state() is True:
                return o.get_label()


def abbreviate_group_name(name: str) -> str:
    if name.startswith("/aws/lambda/"):
        return name.replace("/aws/lambda/", "λ ")
    if name.startswith("/aws/vendedlogs/states/"):
        return name.replace("/aws/vendedlogs/states/", "SF ")
    if name.startswith("API-Gateway-Execution-Logs_"):
        return name.replace("API-Gateway-Execution-Logs_", "API-Exec-")
    return name


def unmask_group_name(name: str) -> str:
    if name.startswith("λ "):
        return name.replace("λ ", "/aws/lambda/")
    if name.startswith("SF "):
        return name.replace("SF ", "/aws/vendedlogs/states/")
    if name.startswith("API-Exec-"):
        return name.replace("API-Exec-", "API-Gateway-Execution-Logs_")
    return name


class GroupSelect(urwid.WidgetWrap):
    def __init__(self, get_selected_region: Callable):
        self.client_map = dict(initial_options())
        self.selected_groups = set()
        self.current_region = (
            get_selected_region() or SETTINGS["presets"]["default"]["region"]
        )
        self.get_selected_region = get_selected_region
        urwid.WidgetWrap.__init__(
            self, self.build_group_checkboxes(self.current_region)
        )

    def toggle_group_select(self, widget, active: bool):
        log.info(
            "log group selected",
            value=active,
            group_name=unmask_group_name(widget.get_label()),
        )
        if active:
            self.selected_groups.add(unmask_group_name(widget.get_label()))
        else:
            self.selected_groups.remove(unmask_group_name(widget.get_label()))

    def get_state(self):
        return list(self.selected_groups)

    def change_region(self, region_name):
        if region_name == self.current_region:
            return
        # clear out group selections since we are swapping region
        self.selected_groups = set()
        self.current_region = region_name
        log.info("loading region groups", region=region_name)
        self._w = self.build_group_checkboxes(region_name)
        log.info("region swap done", region=region_name)

    def build_group_checkboxes(self, region_name: str):
        return urwid.ListBox(
            [
                urwid.CheckBox(
                    abbreviate_group_name(g["name"]),
                    on_state_change=self.toggle_group_select,
                )
                for g in self.client_map[region_name]()
            ]
        )


def loop():
    reader = LogReadingPane()
    loop = urwid.MainLoop(
        reader.top,
        {
            ("timestamp", "dark red", "light gray"),
        },
        unhandled_input=d_widget(reader),
    )
    loop.run()


def initial_options():
    if (USER_DIR / "regions.json").exists():
        with open(USER_DIR / "regions.json") as rjson:
            regions = json.load(rjson)
    else:
        regions = boto3.Session().get_available_regions("logs")
        with open(USER_DIR / "regions.json", "w") as rjson:
            json.dump(regions, rjson)
    for r in regions:
        yield (r, groups_in_region(r))


def groups_in_region(region_name):
    s = boto3.Session(region_name=region_name)
    logs = s.client("logs")

    def regional_logs():
        """Keeps `region_name` inside a closure that will list log groups in that region"""
        try:
            pager = (
                logs.get_paginator("describe_log_groups")
                .paginate()
                .search(
                    "logGroups[].{name: logGroupName, size: storedBytes, created: creationTime}"
                )
            )
            for group in pager:
                since_created = datetime.now(tz=timezone.utc) - datetime.fromtimestamp(
                    group["created"] / 1000, tz=timezone.utc
                )

                if group["size"] or since_created < timedelta(hours=8):
                    group["region"] = region_name
                    yield group
        except logs.exceptions.UnrecognizedClientException:
            pass

    return regional_logs



from typing import List
import os
import time
from datetime import datetime, timedelta, timezone

import boto3

from seesaw import storage
from seesaw.config import logger

log = logger(__file__)


class PolitePoller:
    def __init__(
        self, region: str, logs_client: boto3.client, start_time: datetime = None
    ):
        self.region = region
        self.start_time = start_time or datetime.now(tz=timezone.utc)
        self.logs = logs_client or boto3.client("logs")

    def poller_with_breaks(self, group_name: str, start_time: datetime = None):
        kwargs = {
            "logGroupName": group_name,
            "interleaved": True,
        }
        if start_time is None:
            kwargs["startTime"] = int(self.start_time.timestamp()) * 1000
        else:
            kwargs["startTime"] = int(start_time.timestamp()) * 1000
        while True:
            paginator = self.logs.get_paginator("filter_log_events").paginate(**kwargs)
            log.info("querying", kwargs=kwargs)
            for event in paginator.search("events[]"):
                # update the start time to when we start paginating through
                kwargs["startTime"] = int(
                    max(event["timestamp"] / 1000, kwargs["startTime"])
                )
                event["logGroupName"] = group_name
                event["region"] = self.region
                yield storage.Event.from_json(event)
            yield None


def polling_worker_entry(region, log_group_name):
    session = storage.connect()
    log.info("finding newest message we have")
    newest_msg: List[storage.Event] = (
        session.query(storage.Event)
        .filter(storage.Event.log_group == log_group_name)
        .filter(storage.Event.region == region)
        .order_by(storage.Event.timestamp.desc())
        .limit(1)
        .all()
    )
    if newest_msg:
        start_at = datetime.fromisoformat(newest_msg[0].timestamp)
        log.info(
            "found existing message",
            message_id=newest_msg[0].gid,
            start_at=start_at.isoformat(),
            event_time=newest_msg[0].timestamp,
        )
    else:
        start_at = datetime.now(tz=timezone.utc) - timedelta(hours=4)
    pp = PolitePoller(
        region=region,
        logs_client=boto3.client("logs", region_name=region),
        start_time=start_at,
    )
    session.add(
        storage.LiveReader(
            pid=os.getpid(),
            log_group=log_group_name,
            region=region,
        )
    )
    session.commit()
    log.info("Instantiated poller", region=region, group=log_group_name)
    for msg in pp.poller_with_breaks(group_name=log_group_name):
        if msg is None:
            try:
                session.commit()
                log.info("Saved writes to DB")
            except:
                log.exception("Error committing writes to DB")
            # betweeen message groups, sleep for 10 seconds
            log.info("awaiting more logs", region=region, log_group_name=log_group_name)
            # TODO(rsb): check if we are still supposed to be running
            if storage.LiveReader.exists_for(session, region, log_group_name):
                time.sleep(10)
            else:
                log.info(
                    "self not found in active threads",
                    region=region,
                    log_group_name=log_group_name,
                )
                break
        else:
            try:
                if (
                    session.query(storage.Event)
                    .filter(storage.Event.gid == msg.gid)
                    .one_or_none()
                    is None
                ):
                    session.add(msg)
                    log.info("Added log from", group=msg.log_group, ts=msg.timestamp)
                    session.commit()
            except:
                log.exception("could not add message to DB")


import json
import logging
import os
import pathlib
from pathlib import Path

import appdirs
import structlog
import structlog.contextvars

APP_NAME = "seesaw"
APP_AUTHOR = "ryansb"
VERSION = "0.1.1"

USER_DIR = Path(appdirs.user_config_dir(appname=APP_NAME, appauthor=APP_AUTHOR))
LOG_DIR = Path(
    appdirs.user_log_dir(appname=APP_NAME, appauthor=APP_AUTHOR, version=VERSION)
)

CONF_DEFAULTS = {
    "schema": "0.1",
    "presets": {
        "default": {
            "region": "us-east-1",
            "log_groups": [
                "/aws/apigateway/welcome",
            ],
        },
    },
}

if not LOG_DIR.is_dir():
    LOG_DIR.mkdir(parents=True)
if not USER_DIR.is_dir():
    USER_DIR.mkdir(parents=True)
if not (USER_DIR / "config.json").exists():
    with open(USER_DIR / "config.json", "w") as cfg:
        json.dump(CONF_DEFAULTS, cfg, sort_keys=True, indent=2)

with open(USER_DIR / "config.json", "r") as cfg:
    SETTINGS = json.load(cfg)


root = logging.getLogger()

log_level = logging.DEBUG if os.getenv("SHARE_DEBUG") else logging.INFO

logging.basicConfig(
    format="%(message)s", filename=str(LOG_DIR / "logs.json"), level=log_level
)

logging.getLogger("boto3").setLevel(logging.ERROR)
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)
structlog.configure(
    processors=[
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.processors.format_exc_info,
        structlog.processors.TimeStamper(utc=False, key="timestamp", fmt="iso"),
        structlog.contextvars.merge_contextvars,
        structlog.processors.JSONRenderer(),
    ],
    wrapper_class=structlog.BoundLogger,
    logger_factory=structlog.stdlib.LoggerFactory(),
)

_log_cache = {}


def logger(calling_file: str, context: dict = None) -> structlog.BoundLogger:
    """usage: logger(__file__, {})"""
    path = pathlib.Path(calling_file).parts
    try:
        logger_path = ".".join(path[path.index("seesaw") + 1 :]).replace(".py", "")
    except ValueError:
        # called from outside our package
        logger_path = calling_file.replace("/", ".").replace(".py", "")
    if logger_path not in _log_cache:
        _log_cache[logger_path] = structlog.get_logger(logger_path, **(context or {}))
        return _log_cache[logger_path]
    if context:
        _log_cache[logger_path] = _log_cache[logger_path].new(**context)
    return _log_cache[logger_path]



import os
from datetime import datetime, timezone

import psutil
import sqlalchemy
from sqlalchemy import create_engine

from seesaw.config import USER_DIR, logger

log = logger(__file__)

from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker as _sessionmaker

Base = declarative_base()


def connect() -> sqlalchemy.orm.Session:
    log_db_file = "sqlite:///" + str(USER_DIR / "logs.sqlite")
    log.debug("Opening DB", path=log_db_file)
    engine = create_engine(log_db_file)

    Base.metadata.create_all(engine)

    Session = scoped_session(_sessionmaker())
    Session.configure(bind=engine)
    return Session()


class Event(Base):
    __tablename__ = "events"
    gid = Column(String, primary_key=True)
    timestamp = Column(String, nullable=False)
    ingestion_time = Column(String)
    message = Column(String, default="")
    region = Column(String, nullable=False)
    log_group = Column(String, nullable=False)
    log_stream = Column(String, nullable=False)

    @classmethod
    def from_json(cls, incoming):
        return Event(
            gid=incoming["eventId"],
            timestamp=datetime.fromtimestamp(
                incoming["timestamp"] / 1000, tz=timezone.utc
            ).isoformat(),
            ingestion_time=datetime.fromtimestamp(
                incoming["ingestionTime"] / 1000, tz=timezone.utc
            ).isoformat(),
            region=incoming["region"],
            message=incoming["message"],
            log_stream=incoming["logStreamName"],
            log_group=incoming["logGroupName"],
        )


def get_pid_start_iso() -> str:
    """Return the iso-formatted time the current process started.

    This helps us clear out old thread signatures"""
    pid_start_ts = psutil.Process(os.getpid()).create_time()
    return (
        datetime.utcfromtimestamp(pid_start_ts).replace(tzinfo=timezone.utc).isoformat()
    )


class LiveReader(Base):
    __tablename__ = "live_readers"
    pkey = Column(Integer, autoincrement=True, primary_key=True)
    pid = Column(String)
    log_group = Column(String)
    region = Column(String)
    initiated = Column(String, default=get_pid_start_iso)

    @classmethod
    def exists_for(cls, session: sqlalchemy.orm.Session, region: str, log_group: str):
        return (
            session.query(LiveReader)
            .filter(LiveReader.pid == os.getpid())
            .filter(LiveReader.region == region)
            .filter(LiveReader.log_group == log_group)
            .all()
        )

    @classmethod
    def destroy_other_region_readers(cls, session: sqlalchemy.orm.Session, region: str):
        for reader in (
            session.query(LiveReader)
            .filter(LiveReader.pid == os.getpid())
            .filter(LiveReader.region != region)
            .filter(LiveReader.initiated >= get_pid_start_iso())
            .all()
        ):
            session.delete(reader)
        session.commit()
